/*
 * $Id: builtin.c,v 1.5 2000/09/07 07:10:06 keithp Exp $
 *
 * Copyright  2000 Keith Packard
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Keith Packard not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  Keith Packard makes no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * KEITH PACKARD DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL KEITH PACKARD BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

#include    "formula.h"
#include    <MathLib.h>

typedef  enum _builtins {
    BuiltinLog,
    BuiltinExp,
    BuiltinSin,
    BuiltinCos,
    BuiltinTan,
    BuiltinAsin,
    BuiltinAcos,
    BuiltinAtan,
    BuiltinRandom,
    BuiltinSRandom,
    BuiltinFloor,
    BuiltinCeil,
    BuiltinFrexp,
    BuiltinLdexp,
} BuiltinId;

FormBuiltin builtins[] = {
    { "log",	1,  1,  BuiltinLog },
    { "exp",	1,  1,  BuiltinExp },
    { "sin",	1,  1,	BuiltinSin },
    { "cos",	1,  1,	BuiltinCos },
    { "tan",	1,  1,	BuiltinTan },
    { "asin",	1,  1,	BuiltinAsin },
    { "acos",	1,  1,	BuiltinAcos },
    { "atan",	1,  1,	BuiltinAtan },
    { "random",	0,  1,	BuiltinRandom },
    { "srandom",1,  0,	BuiltinSRandom },
    { "floor",	1,  1,	BuiltinFloor },
    { "ceil",	1,  1,	BuiltinCeil },
    { "frexp",	1,  2,	BuiltinFrexp },
    { "ldexp",	2,  1,	BuiltinLdexp },
};

#define NUM_BUILTIN (sizeof (builtins) / sizeof (builtins[0]))

FormBuiltin *
BuiltinLookup (Compile *c, FormString name)
{
    char    *n = CompString (c, name);
    int	    i;

    for (i = 0; i < NUM_BUILTIN; i++)
	if (!StrCompare (n, builtins[i].name))
	    return &builtins[i];
    return 0;
}

void
BuiltinExecute (FormMachine *m, FormInt id)
{
    Int	    i;
    
    switch (id) {
    case BuiltinLog:
	m->a = log (m->a);
	break;
    case BuiltinExp:
	m->a = exp (m->a);
	break;
    case BuiltinSin:
	m->a = sin (m->a);
	break;
    case BuiltinCos:
	m->a = cos (m->a);
	break;
    case BuiltinTan:
	m->a = tan (m->a);
	break;
    case BuiltinAsin:
	m->a = asin (m->a);
	break;
    case BuiltinAcos:
	m->a = acos (m->a);
	break;
    case BuiltinAtan:
	m->a = atan (m->a);
	break;
    case BuiltinRandom:
	m->a = ((FormNumber) SysRandom (0)) / ((FormNumber) sysRandomMax);
	break;
    case BuiltinSRandom:
	(void) SysRandom ((Int) (m->a * ((FormNumber) sysRandomMax)));
	break;
    case BuiltinFloor:
	m->a = floor (m->a);
	break;
    case BuiltinCeil:
	m->a = ceil (m->a);
	break;
    case BuiltinFrexp:
	m->a = frexp (m->a, &i);
	ValueCheckPush (m, ((FormNumber) i));
	break;
    case BuiltinLdexp:
	i = (int) ValuePop (m);
	m->a = ldexp (m->a, i);
	break;
    default:
	MachException (m, "Undefined builtin function", 0);
	break;
    }
}
