/*
 * $Id: compile.c,v 1.13 2000/09/07 07:10:06 keithp Exp $
 *
 * Copyright  2000 Keith Packard
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Keith Packard not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  Keith Packard makes no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * KEITH PACKARD DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL KEITH PACKARD BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

#include "formula.h"

FormInt
CompileByte (Compile *c, FormByte byte)
{
    FormInt e;
    
    e = c->nexe;
    if (c->nexe == c->maxexe)
    {
	if (!ReallocTable (c, &c->exebufHand, c->maxexe + BUF_EXE))
	{
	    yyerror ("overflow object table");
	    return 0;
	}
	c->maxexe += BUF_EXE;
    }
    c->nexe++;
    *CompExe(c, e) = byte;
    return 1;
}

extern void PrintOp (int op);

FormInt
CompileOp (Compile *c, FormOpCode op)
{
    return CompileByte (c, op);
}

FormInt
CompileInt (Compile *c, FormInt i)
{
    CompileByte (c, (FormByte) i);
    return CompileByte (c, (FormByte) (i >> 8));
}

FormInt
CompilePatchInt (Compile *c, FormInt e, FormInt i)
{
    FormByte	*b;

    if (e != -1)
    {
	b = CompExe (c, e);
	b[0] = (FormByte) i;
	b[1] = (FormByte) (i >> 8);
    }
    return e;
}

FormInt
CompileFetchInt (Compile *c, FormInt e)
{
    FormByte	*b;
    if (e != -1)
    {
	b = CompExe (c, e);
	e = b[0] | (b[1] << 8);
    }
    return e;
}

FormInt
CompileAssign (Compile *c, FormInt expr)
{
    Expr    *e;
    FormInt r = 1;

    if (expr == -1)
	return 0;
    e = CompExpr (c, expr);
    switch (e->type) {
    case ExprComma:
	r &= CompileAssign (c, e->u.t.left);
	r &= CompileOp (c, OpPop);
	r &= CompileAssign (c, e->u.t.right);
	break;
    case ExprName:
	r &= CompileOp (c, OpNamePut);
	r &= CompileInt (c, e->u.name);
	break;
    default:
	yyerror ("invalid l-value");
	r = 0;
	break;
    }
    return r;
}

FormInt
CompileInc (Compile *c, Expr *e, Boolean plus, Boolean pre)
{
    FormInt r = 1;
    FormInt one;

    if (plus)
	one = ConstAlloc (c, NumberOne);
    else
	one = ConstAlloc (c, -NumberOne);
    if (one == -1)
	return 0;
    r &= CompileExpr (c, e->u.t.left);
    r &= CompileOp (c, OpPush);
    if (!pre)
        r &= CompileOp (c, OpPush);
    r &= CompileOp (c, OpConst);
    r &= CompileInt (c, one);
    r &= CompileOp (c, OpPlus);
    r &= CompileAssign (c, e->u.t.left);
    if (!pre)
	r &= CompileOp (c, OpPop);
    return r;
}

FormInt
CompileExpr (Compile *c, FormInt expr)
{
    Expr    *e;
    FormInt r = 1;
    
    if (expr == -1)
	return 0;
    e = CompExpr (c, expr);
    switch (e->type) {
    case ExprName:
	r &= CompileOp (c, OpNameGet);
	r &= CompileInt (c, e->u.name);
	break;
    case ExprCheck:
	r &= CompileOp (c, OpNameCheck);
	r &= CompileInt (c, e->u.name);
	break;
    case ExprConst:
	r &= CompileOp (c, OpConst);
	r &= CompileInt (c, e->u.number);
	break;
    case ExprPlus:
    case ExprMinus:
    case ExprTimes:
    case ExprDivide:
    case ExprDiv:
    case ExprMod:
    case ExprPow:
    case ExprLt:
    case ExprLe:
    case ExprEq:
    case ExprNe:
    case ExprGe:
    case ExprGt:
    case ExprLand:
    case ExprLxor:
    case ExprLor:
	r &= CompileExpr (c, e->u.t.left);
	r &= CompileOp (c, OpPush);
	r &= CompileExpr (c, e->u.t.right);
	switch (e->type)
	{
	case ExprPlus:
	    r &= CompileOp (c, OpPlus); break;
	case ExprMinus:
	    r &= CompileOp (c, OpMinus); break;
	case ExprTimes:
	    r &= CompileOp (c, OpTimes); break;
	case ExprDivide:
	    r &= CompileOp (c, OpDivide); break;
	case ExprDiv:
	    r &= CompileOp (c, OpDiv); break;
	case ExprMod:
	    r &= CompileOp (c, OpMod); break;
	case ExprPow:
	    r &= CompileOp (c, OpPow); break;
	case ExprLt:
	    r &= CompileOp (c, OpLt); break;
	case ExprLe:
	    r &= CompileOp (c, OpLe); break;
	case ExprEq:
	    r &= CompileOp (c, OpEq); break;
	case ExprNe:
	    r &= CompileOp (c, OpNe); break;
	case ExprGe:
	    r &= CompileOp (c, OpGe); break;
	case ExprGt:
	    r &= CompileOp (c, OpGt); break;
	case ExprLand:
	    r &= CompileOp (c, OpLand); break;
	case ExprLxor:
	    r &= CompileOp (c, OpLxor); break;
	case ExprLor:
	    r &= CompileOp (c, OpLor); break;
	default:
	}
	break;
    case ExprAnd:
	r &= CompileExpr (c, e->u.t.left);
	NewForwardBranch (c, LabelExpr);
	UseBranch (c, LabelExpr, OpJeq);
	r &= CompileExpr (c, e->u.t.right);
	ForwardBranchDest (c, LabelExpr);
	CleanLabel (c, 1);
	break;
    case ExprOr:
	r &= CompileExpr (c, e->u.t.left);
	NewForwardBranch (c, LabelExpr);
	UseBranch (c, LabelExpr, OpJne);
	r &= CompileExpr (c, e->u.t.right);
	ForwardBranchDest (c, LabelExpr);
	CleanLabel (c, 1);
	break;
    case ExprUminus:
	r &= CompileExpr (c, e->u.t.left);
	r &= CompileOp (c, OpUminus);
	break;
    case ExprNot:
	r &= CompileExpr (c, e->u.t.left);
	r &= CompileOp (c, OpBang);
	break;
    case ExprLnot:
	r &= CompileExpr (c, e->u.t.left);
	r &= CompileOp (c, OpLnot);
	break;
    case ExprBuiltin:
	if (e->u.t.right != -1)
	    r &= CompileExpr (c, e->u.t.right);
	r &= CompileOp (c, OpBuiltin);
	r &= CompileInt (c, CompExpr (c, e->u.t.left)->u.name);
	break;
    case ExprCall:
	if (e->u.t.right != -1)
	{
	    r &= CompileExpr (c, e->u.t.right);
	    r &= CompileOp (c, OpPush);
	}
	r &= CompileOp (c, OpCall);
	r &= CompileInt (c, CompExpr (c, e->u.t.left)->u.name);
	break;
    case ExprComma:
	r &= CompileExpr (c, e->u.t.right);
	r &= CompileOp (c, OpPush);
	r &= CompileExpr (c, e->u.t.left);
	break;
    case ExprAssign:
	r &= CompileExpr (c, e->u.t.right);
	r &= CompileAssign (c, e->u.t.left);
	break;
    case ExprAssignPlus:
    case ExprAssignMinus:
    case ExprAssignTimes:
    case ExprAssignDivide:
    case ExprAssignDiv:
    case ExprAssignMod:
    case ExprAssignLand:
    case ExprAssignLor:
    case ExprAssignPow:
    case ExprAssignLxor:
	r &= CompileExpr (c, e->u.t.left);
	r &= CompileOp (c, OpPush);
	r &= CompileExpr (c, e->u.t.right);
	switch (e->type) {
	case ExprAssignPlus:
	    r &= CompileOp (c, OpPlus); break;
	case ExprAssignMinus:
	    r &= CompileOp (c, OpMinus); break;
	case ExprAssignTimes:
	    r &= CompileOp (c, OpTimes); break;
	case ExprAssignDivide:
	    r &= CompileOp (c, OpDivide); break;
	case ExprAssignDiv:
	    r &= CompileOp (c, OpDiv); break;
	case ExprAssignMod:
	    r &= CompileOp (c, OpMod); break;
	case ExprAssignPow:
	    r &= CompileOp (c, OpPow); break;
	case ExprAssignLand:
	    r &= CompileOp (c, OpLand); break;
	case ExprAssignLxor:
	    r &= CompileOp (c, OpLxor); break;
	case ExprAssignLor:
	    r &= CompileOp (c, OpLor); break;
	default:
	}
	r &= CompileAssign (c, e->u.t.left);
	break;
    case ExprPlusPlusPre:
	r = CompileInc (c, e, true, true);
	break;
    case ExprPlusPlusPost:
	r = CompileInc (c, e, true, false);
	break;
    case ExprMinusMinusPre:
	r = CompileInc (c, e, false, true);
	break;
    case ExprMinusMinusPost:
	r = CompileInc (c, e, false, false);
	break;
    }
    return r;
}
