/*
 * $Id: execute.c,v 1.10 2000/09/07 07:10:06 keithp Exp $
 *
 * Copyright  2000 Keith Packard
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Keith Packard not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  Keith Packard makes no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * KEITH PACKARD DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL KEITH PACKARD BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

#include    "formula.h"

#undef DEBUG
#ifdef DEBUG
#undef DEBUGOP
#endif

FormNumber
MachFetch (FormMachine	*m,
	   FormInt	sym)
{
    FormVar	*v = &m->exe.vars[sym];
    FormNumber	a;
    int		n;

    switch (v->class) {
    case FormClassParam:
    case FormClassLocal:
    case FormClassResult:
	n = m->fp + v->location;
	if (!GetSet (m->value_set, n))
	{
	    MachException (m, "Uninitialized variable ",
			   m->exe.strings + v->name);
	    a = 0;
	    break;
	}
	a = m->value[n];
	break;
    case FormClassGlobal:
	a = m->globals[v->location];
	break;
    default:
	a = 0;
    }
#ifdef DEBUG
    printf ("Fetch %s %e\n", m->exe.strings + v->name, a);
    SysTaskDelay (30);
#endif
    return a;
}

#ifdef DEBUGOP
void
PrintOp (int op)
{
    char	*s = "unknown";
    
    switch (op) 
    {
    case OpNoop: s = "Noop"; break;
    case OpPush: s = "Push"; break;
    case OpPop: s = "Pop"; break;
    case OpJmp: s = "Jmp"; break;
    case OpJne: s = "Jne"; break;
    case OpJeq: s = "Jeq"; break;
    case OpReturn: s = "Return"; break;
    case OpCall: s = "Call"; break;
    case OpNameGet: s = "NameGet"; break;
    case OpNamePut: s = "NamePut"; break;
    case OpConst: s = "Const"; break;
    case OpUminus: s = "Uminus"; break;
    case OpLnot: s = "Lnot"; break;
    case OpBang: s = "Bang"; break;
    case OpFact: s = "Fact"; break;
    case OpPlus: s = "Plus"; break;
    case OpMinus: s = "Minus"; break;
    case OpTimes: s = "Times"; break;
    case OpDivide: s = "Divide"; break;
    case OpDiv: s = "Div"; break;
    case OpMod: s = "Mod"; break;
    case OpPow: s = "Pow"; break;
    case OpLxor: s = "Lxor"; break;
    case OpLand: s = "Land"; break; 
    case OpLor: s = "Lor"; break;
    case OpEq: s = "Eq"; break;
    case OpNe: s = "Ne"; break;
    case OpLt: s = "Lt"; break;
    case OpGt: s = "Gt"; break;
    case OpLe: s = "Le"; break;
    case OpGe: s = "Ge"; break;
    case OpEnd: s = "End"; break;
    default: printf ("illegal %d\n", op); return;
    }
    printf ("%s\n", s);
}

void
PrintInst (FormMachine	*m)
{
    FormByte	*c;
    
    c = ExeCode (&m->exe, m->pc);
    PrintOp (*c);
    SysTaskDelay (60);
}
#endif

void
MachStore (FormMachine	*m,
	   FormInt	sym,
	   FormNumber	a)
{
    FormVar	*v = &m->exe.vars[sym];
    int		n;

#ifdef DEBUG
    printf ("Store %s %e\n", m->exe.strings + v->name, a);
    SysTaskDelay (30);
#endif
    switch (v->class) {
    case FormClassParam:
    case FormClassLocal:
    case FormClassResult:
	n = m->fp + v->location;
	m->value[n] = a;
	SetSet(m->value_set, n);
	break;
    case FormClassGlobal:
	m->globals[v->location] = a;
	break;
    }
}

void
MachUninit (FormMachine *m, FormInt sym)
{
    FormVar	*v = &m->exe.vars[sym];
    int		n;

    switch (v->class) {
    case FormClassParam:
    case FormClassLocal:
    case FormClassResult:
	n = m->fp + v->location;
	ClrSet(m->value_set, n);
	break;
    case FormClassGlobal:
	break;
    }
}

int
MachIsInit (FormMachine *m, FormInt sym)
{
    FormVar	*v = &m->exe.vars[sym];
    int		n;

    switch (v->class) {
    case FormClassParam:
    case FormClassLocal:
    case FormClassResult:
	n = m->fp + v->location;
	return GetSet(m->value_set, n);
    case FormClassGlobal:
	break;
    }
    return 1;
}

void
MachException (FormMachine *m, char *message, char *optional)
{
    FormInt	func;
    FormFunc	*f;

    if (!optional)
	optional = "";
    for (func = 0; func < m->exe.numFunc; func++)
    {
	f = &m->exe.funcs[func];
	if (f->codeBlock <= m->pc &&
	    m->pc < f->codeBlock + f->codeSize)
	{
	    if (f->label == -1)
	    {
		FormError ("%s%s in %s", message, optional, 
			   ExeString (&m->exe, f->name));
	    }
	    else
	    {
		FormError ("%s%s in %s \"%s\"",
			   message, optional,
			   ExeString (&m->exe, f->name),
			   ExeString (&m->exe, f->label));
	    }
	    break;
	}
    }
    m->state = FormException;
}

FormInt
MachInline (FormMachine *m)
{
    FormByte	*c;
    FormInt	i;

    c = ExeCode (&m->exe, m->pc);
    i = c[0] | (c[1] << 8);
    m->pc += 2;
    return i;
}

FormInt
MachCall (FormMachine	*m, FormInt func)
{
    FormFunc	*f = ExeFunc (&m->exe, func);
    FormReturn	*r;
    FormInt	n;

    if (m->rp == 0)
    {
	MachException (m, "Call stack overflow", 0);
	return -1;
    }
    n = f->numLocal;
    if (f->numResult != -1)
	n += f->numResult;
    if (m->sp < n)
    {
	MachException (m, "Stack overflow", 0);
	return -1;
    }
    r = &m->ret[--m->rp];
    r->pc = m->pc;
    r->func = func;
    r->fp = m->fp;
    m->fp = m->sp;
    while (n--)
	ValuePushU(m);
    m->pc = f->codeBlock;
    return 0;
}

void
MachReturn (FormMachine *m)
{
    FormReturn	*r;
    FormNumber	*result;
    FormFunc	*f;
    FormInt	nresult;

    r = &m->ret[m->rp++];
    f = ExeFunc (&m->exe, r->func);
    result = &m->value[m->fp];
    m->sp = m->fp + f->numParam;
    m->fp = r->fp;
    m->pc = r->pc;
    if (f->numResult == -1)
    {
	/*
	 * results equivalent to parameters; this
	 * "pushes" them onto the stack
	 */
	m->sp = m->sp - f->numParam;
    }
    else
    {
	nresult = f->numResult;
	while (nresult > 1)
	{
	    nresult--;
	    ValuePush (m, *--result);
	}
	if (nresult)
	{
	    m->a = *--result;
	}
    }
}

FormInt
MachRun (FormMachine	*m)
{
    FormByte	*c;
    FormInt	i;
    Word	x, y;
    Boolean	down;
    
    m->state = FormRunning;
    while (m->state == FormRunning)
    {
#ifdef DEBUGOP
	PrintInst (m);
#endif
	EvtGetPen (&x, &y, &down);
	if (down)
	    MachException (m, "Execution aborted by user", 0);
	c = ExeCode (&m->exe, m->pc);
	m->pc++;
	switch (*c) {
	case OpNoop:
	    break;
	case OpPush:
	    ValueCheckPush (m, m->a);
	    break;
	case OpPop:
	    m->a = ValuePop (m); 
	    break;
	case OpJmp:
	    m->pc = MachInline (m);
	    break;
	case OpJne:
	    i = MachInline (m);
	    if (m->a != NumberZero)
		m->pc = i;
	    break;
	case OpJeq:
	    i = MachInline (m);
	    if (m->a == NumberZero)
		m->pc = i;
	    break;
	case OpCall:
	    MachCall (m, MachInline (m));
	    break;
	case OpBuiltin:
	    BuiltinExecute (m, MachInline (m));
	    break;
	case OpNameGet:
	    m->a = MachFetch (m, MachInline (m));
	    break;
	case OpNamePut:
	    MachStore (m, MachInline (m), m->a);
	    break;
	case OpNameCheck:
	    m->a = MachIsInit (m, MachInline (m)) ? NumberOne : NumberZero;
	    break;
	case OpConst:
	    m->a = m->exe.consts[MachInline (m)];
#ifdef DEBUG
	    printf ("Const %e\n", m->a);
	    SysTaskDelay (30);
#endif
	    break;
	case OpUminus:
	    m->a = -m->a; break;
	case OpLnot:
	    m->a = ~(int) m->a; break;
	case OpBang:
	    m->a = !m->a; break;
	case OpFact:
    /*	m->a = MachFactorial (m->a); */ break; 
	case OpPlus:
	    m->a = ValuePop (m) + m->a; break;
	case OpMinus:
	    m->a = ValuePop (m) - m->a; break;
	case OpTimes:
	    m->a = ValuePop (m) * m->a; break;
	case OpDivide:
	    if (m->a == NumberZero)
		MachException (m, "divide by zero", 0);
	    else
		m->a = ValuePop (m) / m->a; 
	    break;
	case OpDiv:
	    if (m->a == NumberZero)
		MachException (m, "divide by zero", 0);
	    else
		m->a = (int) ValuePop (m) / (int) m->a;
	    break;
	case OpMod:
	    if (m->a == NumberZero)
		MachException (m, "modulus by zero", 0);
	    else
		m->a = (int) ValuePop (m) % (int) m->a;
	    break;
	case OpPow:
	    m->a = FormPow (ValuePop (m), m->a); break;
	case OpLxor:
	    m->a = (int) ValuePop (m) ^ (int) m->a; break;
	case OpLand:
	    m->a = (int) ValuePop (m) & (int) m->a; break;
	case OpLor:
	    m->a = (int) ValuePop (m) | (int) m->a; break;
	case OpEq:
	    m->a = ValuePop (m) == m->a; break;
	case OpNe:
	    m->a = ValuePop (m) != m->a; break;
	case OpLt:
	    m->a = ValuePop (m) < m->a; break;
	case OpGt:
	    m->a = ValuePop (m) > m->a; break;
	case OpLe:
	    m->a = ValuePop (m) <= m->a; break;
	case OpGe:
	    m->a = ValuePop (m) >= m->a; break;
	case OpReturn:
 	case OpEnd:
	    if (m->rp == RET_STACK)
		m->state = FormSuccess;
	    else
		MachReturn (m);
	    break;
	default:
    	    MachException (m, "illegal instruction", 0);
	    break;
	}
    }
    return m->state == FormSuccess;
}

void
MachGlobals (FormMachine *m)
{
    int	    i;
    
    for (i = 0; i < m->exe.numGlobal; i++)
	m->globals[i] = ExeConst (&m->exe, ExeGlobal(&m->exe,i));
}

void
MachInitialize (FormMachine *m, FormFunc *f)
{
    int	    nresult;
    
    MemSet (&m->value_set, sizeof (m->value_set), 0);
    m->pc = f->codeBlock;
    m->rp = RET_STACK;
    m->fp = VALUE_STACK - f->numParam;
    nresult = f->numResult;
    if (nresult == -1)
	nresult = 0;
    m->sp = m->fp - (nresult + f->numLocal);
}
