#pragma pack(2)

#ifdef UNIX
#include    <stdio.h>
#else
#include    <stdio.h>
#include <Common.h>
#include <System/SysAll.h>
#include <UI/UIAll.h>
#endif

#define FormulaDBCreator    'frml'
#define FormulaDBType	    'DATA'
#define FormulaFileCreator  'frml'
#define FormulaVersionNum   1
#define FormulaOpVersion    2

#include <string.h>
    
typedef double    	FormNumber;
typedef unsigned char	FormByte;
typedef short		FormInt;
typedef short		FormString;

#define NumberOne   1.0
#define NumberZero  0.0
#define Number(c)   c

typedef enum _OpCode {
    OpNoop,
    OpPush,
    OpPop,
    /*
     * control flow op codes.  2/4-byte argument: absolute target
     */
    OpJmp,
    OpJmpFar,
    OpJne,
    OpJneFar,
    OpJeq,
    OpJeqFar,
    /*
     * function op codes
     */
    OpReturn,	/* no argument */
    OpCall,	/* 2-byte argument -- index into global table */
    OpBuiltin,	/* 2-byte argument -- index into builtin table */
    /*
     * terminal op codes.  2-byte argument: index into global table
     */
    OpNameGet,
    OpNamePut,
    OpNameCheck,
    OpConst,
    /*
     * operator op codes.  No argument
     */
    OpUminus,
    OpLnot,
    OpBang,
    OpFact,
    OpPlus,
    OpMinus,
    OpTimes,
    OpDivide,
    OpDiv,
    OpMod,
    OpPow,
    OpLxor,
    OpLand, 
    OpLor,
    OpEq,
    OpNe,
    OpLt,
    OpGt,
    OpLe,
    OpGe,
#if 0
    /*
     * array op codes.  2 2-byte arguments: name, number of dimensions
     */
    OpArrayGet,
    OpArrayPut,
#endif
    /*
     * last op code
     */
    OpEnd
} FormOpCode;     

/*
 * Format of executable:
 *
 * Header
 *  name
 *  model   (small, medium, large)
 *  number of variables
 *  number of functions
 *  number of constants
 *  size of code block
 *  size of string table
 * Variables
 *  name
 *  label
 *  class
 *  location
 * Functions
 *  name
 *  label
 *  number parameters
 *  number results
 *  number locals
 *  code size
 *  offset of code block
 * Constants
 * Code Blocks
 *  Parameters
 *   list of variables
 *  Results
 *   list of variables
 *  Locals
 *   list of variables
 *  code
 *   list of bytes
 * Strings
 *   list of bytes; each string is null-terminated
 */

typedef enum _FormClass {
    FormClassParam,
    FormClassResult,
    FormClassLocal,
    FormClassGlobal
} FormClass;

typedef struct _FormVar {
    FormString	name;
    FormString	label;
    FormClass	class;
    FormInt	location;
    FormInt	optional;
#if 0
    FormInt	array;
#endif
} FormVar;

#if 0
typedef struct _FormArray {
    FormInt	dim;
    FormInt	array;
} FormArray;
#endif

typedef struct _FormFunc {
    FormString	name;
    FormString	label;
    FormInt	vars;
    FormInt	numParam;
    FormInt	numResult;
    FormInt	numLocal;
    FormInt	codeSize;
    FormInt	codeBlock;
} FormFunc;

typedef struct _FormBuiltin {
    char	name[16];
    FormInt	numParam;
    FormInt	numResult;
    FormInt	id;
} FormBuiltin;
    
typedef struct _FormExe {
    FormString	name;
    FormInt	numConst;
    FormNumber	*consts;
    FormInt	numGlobal;
    FormInt	*globals;
    FormInt	numVar;
    FormVar	*vars;
#if 0
    FormInt	numArray;
    FormArray	*arrays;
#endif
    FormInt	numFunc;
    FormFunc	*funcs;
    FormByte	*code;
    FormByte	*strings;
} FormExe;

#define ExeString(e,n)	((e)->strings + (n))
#define ExeVar(e,v)	(&(e)->vars[v])
#define ExeArray(e,a)	(&(e)->arrays[a])
#define ExeConst(e,n)	((e)->consts[n])
#define ExeCode(e,n)	((e)->code + (n))
#define ExeFunc(e,f)	(&(e)->funcs[f])
#define ExeGlobal(e,g)	((e)->globals[g])

typedef struct _FormPickle {
    FormString	name;		/* offset from start of pickle */
    FormInt	version;
    FormInt	numConst;
    FormInt	numGlobal;
    FormInt	numVar;
#if 0
    FormInt	numArray;
#endif
    FormInt	numFunc;
    FormInt	codeSize;
    FormInt	stringSize;
    FormInt	totalSize;
} FormPickle;

#define PickleName(p)	(((char *) (p)) + (p)->name)

typedef struct _FormReturn {
    FormInt	pc;
    FormInt	fp;
    FormInt	func;
} FormReturn;

#define VALUE_STACK	256
#define RET_STACK	256
#define GLOBAL_STORE	64

typedef enum _formState {
    FormRunning,
    FormException,
    FormSuccess
} FormState;

#define GetSet(a,i) (((a)[(i)/8] >> ((i) & 7)) & 1)
#define SetSet(a,i) ((a)[(i)/8] |= 1 << ((i) & 7))
#define ClrSet(a,i) ((a)[(i)/8] &= ~(1 << ((i) & 7)))

typedef unsigned char	FormSet;

typedef struct _FormMachine {
    FormExe	exe;
    FormInt	pc;
    FormInt	sp;
    FormInt	fp;
    FormInt	rp;
    FormState	state;
    FormNumber	a;
    FormNumber	globals[GLOBAL_STORE];
    FormNumber	value[VALUE_STACK];
    FormSet	value_set[VALUE_STACK/8];
    FormReturn	ret[RET_STACK];
} FormMachine;

typedef FormByte (*FormReader)(FormInt);

#define ValuePush(m,v)	(((m)->value[--(m)->sp] = (v)), \
			 SetSet((m)->value_set, (m)->sp))
#define ValuePushU(m)	(((m)->value[--(m)->sp] = NumberZero), \
			 ClrSet((m)->value_set, (m)->sp))
#define ValuePop(m)	((m)->value[(m)->sp++])
#define ValueZero(v)	((v) == 0)
#define ValueCheckPush(m,v) {\
    if (m->sp == 0) \
	MachException (m, "Stack overflow", 0);\
    else\
	ValuePush (m, m->a);\
}



typedef enum _exprType {
    ExprName,
    ExprCheck,
    ExprConst,
    ExprPlus,
    ExprMinus,
    ExprTimes,
    ExprDivide,
    ExprDiv,
    ExprMod,
    ExprPow,
    ExprLt,
    ExprLe,
    ExprEq,
    ExprNe,
    ExprGe,
    ExprGt,
    ExprAnd,
    ExprOr,
    ExprNot,
    ExprLand,
    ExprLxor,
    ExprLor,
    ExprLnot,
    ExprComma,
    ExprUminus,
    ExprCall,
    ExprBuiltin,
    ExprAssign,
    ExprAssignPlus,
    ExprAssignMinus,
    ExprAssignTimes,
    ExprAssignDivide,
    ExprAssignDiv,
    ExprAssignMod,
    ExprAssignLand,
    ExprAssignLor,
    ExprAssignPow,
    ExprAssignLxor,
    ExprPlusPlusPre,
    ExprPlusPlusPost,
    ExprMinusMinusPre,
    ExprMinusMinusPost,
#if 0
    ExprArray
#endif
} ExprType;

typedef struct _expr {
    ExprType	type;
    FormInt	nvalues;
    union {
	struct {
	    FormInt	left;
	    FormInt	right;
	} t;
	FormInt		name;
#if 0
	struct {
	    FormInt	name;
	    FormInt	array;
	} n;
#endif
	FormInt number;
	FormInt	builtin;
    } u;
} Expr;

typedef struct _name {
    FormInt	scopedOut;
    FormVar	var;
} Name;
    
typedef enum _labelType {
    LabelBreak,
    LabelContinue,
    LabelIfFalse,
    LabelEndif,
    LabelDo,
    LabelExpr
} LabelType;

typedef struct _label {
    FormInt	exe;
    FormInt	reverse;
    LabelType	type;
} Label;
    
#define NUM_VAR	    16
#define NUM_ARRAY   8
#define NUM_GLOBAL  4
#define NUM_CONST   4
#define NUM_EXPR    32
#define NUM_FUNC    4
#define NUM_STRING  16
#define NUM_LABEL   16
#define BUF_STRING  128
#define BUF_EXE	    256
#define BUF_ERROR   256

typedef struct _compile {
    FormString	name;
    int		nvar;
    int		maxvar;
    Name	*var;
    VoidHand	varHand;
#if 0
    int		narray;
    int		maxarray;
    FormArray	*array;
    VoidHand	arrayHand;
#endif
    int		nglobal;
    int		maxglobal;
    FormInt	*globals;
    VoidHand	globalsHand;
    int		nconst;
    int		maxconst;
    FormNumber	*consts;
    VoidHand	constsHand;
    int		nexpr;
    int		maxexpr;
    Expr	*expr;
    VoidHand	exprHand;
    int		nfunc;
    int		maxfunc;
    FormFunc	*funcs;
    VoidHand	funcsHand;
    int		nstring;
    int		maxstring;
    FormInt	*strings;
    VoidHand	stringsHand;
    int		nstringbuf;
    int		maxstringbuf;
    FormByte	*stringbuf;
    VoidHand	stringbufHand;
    int		nlabel;
    int		maxlabel;
    Label	*labels;
    VoidHand	labelsHand;
    int		nexe;
    int		maxexe;
    FormByte	*exebuf;
    VoidHand	exebufHand;
    char	error[BUF_ERROR];
} Compile;

extern Compile	*compile;

#define CompExpr(c,e)	(&(c)->expr[e])
#define CompVar(c,v)	(&(c)->var[v])
#define CompArray(c,a)	(&(c)->array[a])
#define CompGlobal(c,g)	(&(c)->globals[g])
#define CompConst(c,i)	(&(c)->consts[i])
#define CompFunc(c,f)	(&(c)->funcs[f])
#define CompLabel(c,l)	(&(c)->labels[l])
#define CompExe(c,e)	(&(c)->exebuf[e])
#define CompString(c,s)	((c)->stringbuf + (s))

/*
 * builtin.c
 */

FormBuiltin *BuiltinLookup (Compile *, FormString);
void	    BuiltinExecute (FormMachine *m, FormInt id);

 
/*
 * compile.c
 */

FormInt	CompileByte (Compile *, FormByte byte);
FormInt	CompileOp (Compile *c, FormOpCode op);
FormInt	CompileInt (Compile *c, FormInt i);
FormInt	CompilePatchInt (Compile *c, FormInt e, FormInt i);
FormInt	CompileFetchInt (Compile *c, FormInt e);
FormInt	CompileExpr (Compile *, FormInt e);

/*
 * execute.c
 */

void	MachStore (FormMachine *m, FormInt sym, FormNumber a);
FormNumber  MachFetch (FormMachine *m, FormInt sym);
void	MachUninit (FormMachine *m, FormInt sym);
int	MachIsInit (FormMachine *m, FormInt sym);
void	MachInitialize (FormMachine *m, FormFunc *f);
void	MachGlobals (FormMachine *m);
FormInt	MachRun (FormMachine *);
void	MachException (FormMachine *, char *, char *);

/*
 * formula.c
 */

void	FormError (char *fmt, ...);
void	vsprint (char *dst, char *format, va_list args);

/*
 * gram.y
 */

void	yyinit (char *program, int programSize);
int	yyparse (void);
int	yywrap (void);
void	yyerror (char *fmt, ...);
void	yysprint (char *dst, char *format, ...);
FormPickle	*FormParse (char *program, int programSize);

/*
 * lex.c
 */
#define LEX_TOKEN   128

int	lex (unsigned char *);
void	lexinit (unsigned char *program, int programSize);
int	atodouble (char *, FormNumber *);
void	doubletoa (FormNumber, char *);
void	inttoa (int, char *);
unsigned char	*lexgetline (int *);

extern FormNumber   yydval;

/*
 * load.c
 */
FormInt		FormLoad (FormPickle *p, FormExe *e);
FormPickle	*FormPack (Compile *c);

/*
 * main.c
 */

int main (int argc, char **argv);

/*
 * node.c
 */

FormInt	ExprAlloc (Compile *c, ExprType type);
FormInt	ExprBinary (Compile *c, ExprType type, FormInt left, FormInt right);
FormInt	ExprUnary (Compile *c, ExprType type, FormInt down);
FormInt	ExprTree (Compile *c, ExprType type, FormInt left, FormInt right);
FormInt ExprVar (Compile *c, FormInt name);
FormInt ExprVarCheck (Compile *c, FormInt name);
FormInt ExprBuilt (Compile *c, FormInt id);
FormInt ExprNumber (Compile *c, FormInt number);
FormInt	ConstAlloc (Compile *c, FormNumber n);
FormInt	NameAlloc (Compile *c, FormInt name, FormInt optional, FormInt label, FormClass class);
FormInt	NameLookup (Compile *c, FormInt name);
FormInt	FuncAlloc (Compile *c, FormInt name);
FormInt	FuncLookup (Compile *c, FormInt name);
FormInt	StringAlloc (Compile *c, FormByte *string);
FormInt	GlobalAlloc (Compile *c, FormInt number);
void	CleanLocals (Compile *c);
FormInt	CountClass (Compile *c, FormClass class, FormInt base);
FormInt	PushLabel (Compile *c, LabelType type, FormInt e);
FormInt	NewForwardBranch (Compile *c, LabelType type);
FormInt	UseBranch (Compile *c, LabelType type, FormOpCode op);
FormInt	ForwardBranchDest (Compile *c, LabelType type);
FormInt	ReverseBranchDest (Compile *c, LabelType type);
void	CleanLabel (Compile *c, FormInt n);
FormInt	ReallocTable (Compile *c, VoidHand *h, FormInt size);
void	FreeHandles (Compile *c);
void	FreeTemp (Compile *c);

/*
 * prim.c
 */

FormNumber  FormPow (FormNumber a, FormNumber b);
