%{
/*
 * $Id: gram.y,v 1.13 2000/09/07 09:16:18 keithp Exp $
 *
 * Copyright  2000 Keith Packard
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Keith Packard not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  Keith Packard makes no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * KEITH PACKARD DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL KEITH PACKARD BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

#include "formula.h"
#define alloca	really_bad_dont_do_this

#define yyoverflow(m,a1,a2,a3,a4,a5) { yyerror ("parser stack overflow"); return 2; }
  
static int (*lexp)(unsigned char *) = lex;
static void (*lexinitp)(unsigned char *, int) = lexinit;
static void (*formError)(char *,...) = FormError;
static FormBuiltin *(*builtinLookup)(Compile *, FormString) = BuiltinLookup;
static int  yylex (void);

#define yylexinit (*lexinitp)

#if 0
static int  (*printfp)(const char *, ...) = printf;

#define printf (*printfp)
#endif

FormClass   class;
FormInt	    func;
FormInt	    funcNumParam, funcNumResult;
Compile	    *compile;
%}
%token  ERROR
%token	NAME LABEL NUMBER
%token	BREAK CONTINUE DECLARE DEFAULT DEFINE DO ELSE FOR IF LOCAL PRIVATE RETURN SWITCH WHILE
%token	';' '(' ')' '{' '}' ','
%right	'=' MODEQ LANDEQ TIMESEQ PLUSEQ MINUSEQ DIVIDEEQ DIVEQ LXOREQ POWEQ LOREQ
%left	','
%left	OR
%left	AND
%left	'|'
%left	LXOR
%left	'&'
%left	EQ NE
%left   '<' '>' GE LE
%left	'+' '-'
%left	'*' '/' '%' DIV
%right	'^'
%right	UMINUS '!' '~' PLUSPLUS MINUSMINUS
%%
file	:   oplabel formulas
		{ compile->name = $1; }
	;
formulas:   formula formulas
	|
	;
formula	:   global
	|   function
	|   declare
	;
global	:   NAME '=' NUMBER ';'
		{
		    FormInt n, g;

		    if ($1 == -1)
			YYERROR;
		    if ($3 == -1)
			YYERROR;
		    n = NameAlloc (compile, $1, 0, -1, FormClassGlobal);
		    if (n < 0)
		    {
			YYERROR;
		    }
		    g = GlobalAlloc (compile, $3);
		    if (g < 0)
		    {
			YYERROR;
		    }
		    CompVar (compile, n)->var.location = g;
		}
	;
function:   DEFINE oprivate NAME oplabel
		{
		    FormFunc    *f;

		    if ($3 == -1)
			YYERROR;
		    func = FuncLookup (compile, $3);
		    if (func != -1)
		    {
			f = CompFunc (compile, func);
			if (f->vars != -1)
			{
			    yyerror ("redefining \"%s\"",
				     CompString (compile, $3));
			    YYERROR;
			}
			funcNumParam = f->numParam;
			funcNumResult = f->numResult;
		    } 
		    else
		    {
			func = FuncAlloc (compile, $3);
			if (func < 0)
			    YYERROR;
			f = CompFunc (compile, func);
			funcNumParam = -2;
			funcNumResult = -2;
		    }
		    if ($2)
			f->label = -2;
		    else
			f->label = $4;
		    f->codeBlock = compile->nexe;
		    f->vars = compile->nvar;
		}
	    outputs inputs '{' locals stats '}'
		{
		    FormFunc    *f;
		    FormInt	n, paramo, localo, resulto;
		    Name	*name;
		    
		    if (!CompileOp (compile, OpEnd))
		    {
			yyerror ("overflow object table");
			YYERROR;
		    }
		    f = CompFunc (compile, func);
		    if (!$6)
		    {
			f->numResult = -1;
		    }
		    if (funcNumResult != -2)
		    {
			if (f->numResult != funcNumResult ||
			    f->numParam != funcNumParam)
			{
			    yyerror ("mismatch declaration \"%s\"",
				     CompString (compile, $3));
			    YYERROR;
			}
		    }
		    f->codeSize = compile->nexe - f->codeBlock;
		    paramo = 0;
		    if (f->numResult == -1)
		    {
			resulto = 0;
			localo = -f->numLocal;
		    }
		    else
		    {
			resulto = -f->numResult;
			localo = -(f->numResult + f->numLocal);
		    }
		    for (n = f->vars; n < compile->nvar; n++)
		    {
			name = CompVar (compile, n);
			switch (name->var.class) {
			case FormClassParam:
			    name->var.location = paramo++;
			    break;
			case FormClassResult:
			    name->var.location = resulto++;
			    break;
			case FormClassLocal:
			    name->var.location = localo++;
			    break;
			case FormClassGlobal:
			}
		    }
		    CleanLocals (compile);
		}
	;
oprivate:   PRIVATE
		{ $$ = 1; }
	|
		{ $$ = 0; }
	;
outputs	:   { class = FormClassResult; } '(' odecls ')'
		{ $$ = 1; }
	|   '*'
		{ $$ = 0; }
	;
inputs	:   { class = FormClassParam; } '(' odecls ')'
	;
odecls	:  decls
	|
	;
decls	:   decl ',' decls
	|   decl
	;
decl	:   NAME option oplabel
		{   FormInt   n;
		    if ($1 == -1)
		    {
			yyerror ("overflow string table");
			YYERROR;
		    }
		    n = NameAlloc (compile, $1, $2, $3, class);
		    if (n < 0)
		    {
			if (n == -1)
			{
			    yyerror ("overflow var table");
			}
			else
			{
			    yyerror ("redefining \"%s\"",
				     CompString (compile, $1));
			}
			YYERROR;
		    }
		    switch (class) {
		    case FormClassParam:
			CompFunc(compile, func)->numParam++;
			break;
		    case FormClassResult:
			CompFunc(compile, func)->numResult++;
			break;
		    case FormClassLocal:
			CompFunc(compile, func)->numLocal++;
			break;
		    case FormClassGlobal:
			break;
		    }
		}
	;
option	:   '?'
	    { $$ = 1; }
	|
	    { $$ = 0; }
	;
declare	:   DECLARE NAME doutputs dinputs ';'
		{
		    FormFunc    *f;

		    if ($2 == -1)
		    {
			yyerror ("overflow string table");
			YYERROR;
		    }
		    func = FuncLookup (compile, $2);
		    if (func != -1)
		    {
			f = CompFunc (compile, func);
			if (f->numParam != $4 || f->numResult != $3)
			{
			    yyerror ("mismatch definition \"%s\"",
				     CompString (compile, $2));
			    YYERROR;
			}
		    }
		    else
		    {
			func = FuncAlloc (compile, $2);
			if (func == -1)
			    YYERROR;
			f = CompFunc (compile, func);
			f->numParam = $4;
			f->numResult = $3;
			f->vars = -1;
		    }
		}
	;
doutputs:   '(' oddecls ')'
		{ $$ = $2; }
	|   '*'
		{ $$ = -1; }
	;
dinputs	:   '(' oddecls ')'
		{ $$ = $2; }
	;
oddecls	:  ddecls
	|
		{ $$ = 0; }
	;
ddecls	:   ddecl ',' ddecls
		{ $$ = $3 + 1; }
	|   ddecl
		{ $$ = 1; }
	;
ddecl	:   NAME oplabel
		{
		    if ($1 == -1)
			YYERROR;
		    $$ = $1;
		}
	;
oplabel	:   LABEL
		{
		    if ($1 == -1)
			YYERROR;
		    $$ = $1;
		}
	|
		{ $$ = -1; }
	;
locals	:   LOCAL { class = FormClassLocal; } decls ';'
	|
	;
stats	:   stat stats
	|
	;
stat	:   WHILE '(' expr ')' 
		{ 
		    if (ReverseBranchDest (compile, LabelContinue) == -1)
			YYERROR;
		    if (!CompileExpr (compile, $3))
			YYERROR;
		    if (NewForwardBranch (compile, LabelBreak) == -1)
			YYERROR;
		    UseBranch (compile, LabelBreak, OpJeq);
		    compile->nexpr = 0;
		}
	    stat
		{ 
		    UseBranch (compile, LabelContinue, OpJmp);
		    ForwardBranchDest (compile, LabelBreak);
		    CleanLabel (compile, 2);
		}
	|   IF '(' expr ')' 
		{
		    if (NewForwardBranch (compile, LabelIfFalse) == -1)
			YYERROR;
		    if (!CompileExpr (compile, $3))
			YYERROR;
		    UseBranch (compile, LabelIfFalse, OpJeq);
		    compile->nexpr = 0;
		}
	    stat oelse
	|   DO 
		{
		    if (ReverseBranchDest (compile, LabelDo) == -1)
			YYERROR;
		    if (NewForwardBranch (compile, LabelBreak) == -1)
			YYERROR;
		    if (NewForwardBranch (compile, LabelContinue) == -1)
			YYERROR;
		}
	    stat WHILE '(' expr ')' ';'
		{
		    ForwardBranchDest (compile, LabelContinue);
		    if (!CompileExpr (compile, $6))
			YYERROR;
		    UseBranch (compile, LabelDo, OpJne);
		    ForwardBranchDest (compile, LabelBreak);
		    CleanLabel (compile, 2);
		    compile->nexpr = 0;
		}
	|   FOR '(' oexpr ';' oexpr ';' oexpr ')' 
		{
		    if ($3 != -1)
		    {
			if (!CompileExpr (compile, $3))
			    YYERROR;
		    }
		    
		    if (NewForwardBranch (compile, LabelDo) == -1)
			YYERROR;
		    if ($7 != -1)
			UseBranch (compile, LabelDo, OpJmp);
		    if (ReverseBranchDest (compile, LabelContinue) == -1)
			YYERROR;
		    if ($7 != -1)
		    {
			if (!CompileExpr (compile, $7))
			    YYERROR;
		    }
		    ForwardBranchDest (compile, LabelDo);
		    
		    if (NewForwardBranch (compile, LabelBreak) == -1)
			YYERROR;
		    if ($5 != -1)
		    {
			if (!CompileExpr (compile, $5))
			    YYERROR;
			UseBranch (compile, LabelBreak, OpJeq);
		    }
		    compile->nexpr = 0;
		}
	    stat
		{
		    UseBranch (compile, LabelContinue, OpJmp);
		    ForwardBranchDest (compile, LabelBreak);
		    CleanLabel (compile, 3);
		}
	|   RETURN ';'
		{
		    if (!CompileOp (compile, OpReturn))
			YYERROR;
		}
	|   BREAK ';'
		{
		    if (UseBranch (compile, LabelBreak, OpJmp) == -1)
			YYERROR;
		}
	|   CONTINUE ';'
		{
		    if (UseBranch (compile, LabelContinue, OpJmp) == -1)
			YYERROR;
		}
	|   expr ';'
		{
		    FormInt	npop;
		    if (!CompileExpr (compile, $1))
			YYERROR;
		    npop = CompExpr (compile, $1)->nvalues;
		    while (npop > 1)
		    {
			if (!CompileOp (compile, OpPop))
			    YYERROR;
			npop--;
		    }
		    compile->nexpr = 0;
		}
	|   ';'
	|   '{' stats '}'
	;
oelse	:   ELSE 
		{
		    if (NewForwardBranch (compile, LabelEndif) == -1)
			YYERROR;
		    UseBranch (compile, LabelEndif, OpJmp);
		    ForwardBranchDest (compile, LabelIfFalse);
		}
	    stat
		{
		    ForwardBranchDest (compile, LabelEndif);
		    CleanLabel (compile, 2);
		}

	|
		{
		    ForwardBranchDest (compile, LabelIfFalse);
		    CleanLabel (compile, 1);
		}
	;
var	:   NAME
		{
		    FormInt v;

		    if ($1 == -1)
			YYERROR;
		    v = NameLookup (compile, $1);
		    if (v == -1)
		    {
			yyerror ("undefined \"%s\"",
				 CompString (compile, $1));
			YYERROR;
		    }
		    $$ = ExprVar (compile, v);
		    if ($$ == -1) YYERROR;
		}
	;
assignop:   '='
		{ $$ = ExprAssign; }
	|   PLUSEQ
		{ $$ = ExprAssignPlus; }
	|   MINUSEQ
		{ $$ = ExprAssignMinus; }
	|   TIMESEQ
		{ $$ = ExprAssignTimes; }
	|   DIVIDEEQ
		{ $$ = ExprAssignDivide; }
	|   DIVEQ
		{ $$ = ExprAssignDiv; }
	|   MODEQ
		{ $$ = ExprAssignMod; }
	|   LANDEQ
		{ $$ = ExprAssignLand; }
	|   LOREQ
		{ $$ = ExprAssignLor; }
	|   POWEQ
		{ $$ = ExprAssignPow; }
	|   LXOREQ
		{ $$ = ExprAssignLxor; }
	;
expr	:   var
	|   expr assignop expr	%prec '='
		{
		    if (CompExpr (compile, $1)->nvalues !=
			CompExpr (compile, $3)->nvalues)
		    {
			yyerror ("Can't assign %d values to %d variables",
				 CompExpr (compile, $3)->nvalues,
				 CompExpr (compile, $1)->nvalues);
			YYERROR;
		    }
		    if ($2 != ExprAssign && CompExpr (compile, $1)->nvalues != 1)
		    {
			yyerror ("Can't operator-assign more than one value");
			YYERROR;
		    }
		    $$ = ExprTree (compile, $2,
				   $1, $3);
		    if ($$ == -1) YYERROR;
		    CompExpr (compile, $$)->nvalues = 1;
		}
	|   var PLUSPLUS
		{ $$ = ExprUnary (compile, ExprPlusPlusPost, $1); if ($$ == -1) YYERROR; }
	|   var MINUSMINUS
		{ $$ = ExprUnary (compile, ExprMinusMinusPost, $1); if ($$ == -1) YYERROR; }
	|   PLUSPLUS var
		{ $$ = ExprUnary (compile, ExprPlusPlusPre, $2); if ($$ == -1) YYERROR; }
	|   MINUSMINUS var
		{ $$ = ExprUnary (compile, ExprMinusMinusPost, $2); if ($$ == -1) YYERROR; }
	|   '?' NAME
		{
		    FormInt v;

		    if ($2 == -1)
			YYERROR;
		    v = NameLookup (compile, $2);
		    if (v == -1)
		    {
			yyerror ("undefined \"%s\"",
				 CompString (compile, $2));
			YYERROR;
		    }
		    $$ = ExprVarCheck (compile, v);
		    if ($$ == -1) YYERROR;
		}
	|   NUMBER
		{ 
		    if ($1 == -1)
			YYERROR;
		    $$ = ExprNumber (compile, $1); if ($$ == -1) YYERROR; 
		}
	|   expr '+' expr
		{ $$ = ExprBinary (compile, ExprPlus, $1, $3); if ($$ == -1) YYERROR; }
	|   expr '-' expr
		{ $$ = ExprBinary (compile, ExprMinus, $1, $3); if ($$ == -1) YYERROR; }
	|   expr '*' expr
		{ $$ = ExprBinary (compile, ExprTimes, $1, $3); if ($$ == -1) YYERROR; }
	|   expr '/' expr
		{ $$ = ExprBinary (compile, ExprDivide, $1, $3); if ($$ == -1) YYERROR; }
    	|   expr DIV expr
		{ $$ = ExprBinary (compile, ExprDiv, $1, $3); if ($$ == -1) YYERROR; }
	|   expr '%' expr
		{ $$ = ExprBinary (compile, ExprMod, $1, $3); if ($$ == -1) YYERROR; }
	|   expr '^' expr
		{ $$ = ExprBinary (compile, ExprPow, $1, $3); if ($$ == -1) YYERROR; }
	|   expr '<' expr
		{ $$ = ExprBinary (compile, ExprLt, $1, $3); if ($$ == -1) YYERROR; }
	|   expr LE expr
		{ $$ = ExprBinary (compile, ExprLe, $1, $3); if ($$ == -1) YYERROR; }
	|   expr EQ expr
		{ $$ = ExprBinary (compile, ExprEq, $1, $3); if ($$ == -1) YYERROR; }
	|   expr NE expr
		{ $$ = ExprBinary (compile, ExprNe, $1, $3); if ($$ == -1) YYERROR; }
	|   expr GE expr
		{ $$ = ExprBinary (compile, ExprGe, $1, $3); if ($$ == -1) YYERROR; }
	|   expr '>' expr
		{ $$ = ExprBinary (compile, ExprGt, $1, $3); if ($$ == -1) YYERROR; }
	|   expr AND expr
		{ $$ = ExprBinary (compile, ExprAnd, $1, $3); if ($$ == -1) YYERROR; }
	|   expr OR expr
		{ $$ = ExprBinary (compile, ExprOr, $1, $3); if ($$ == -1) YYERROR; }
	|   expr '&' expr
		{ $$ = ExprBinary (compile, ExprLand, $1, $3); if ($$ == -1) YYERROR; }
	|   expr LXOR expr
		{ $$ = ExprBinary (compile, ExprLxor, $1, $3); if ($$ == -1) YYERROR; }
	|   expr '|' expr
		{ $$ = ExprBinary (compile, ExprLor, $1, $3); if ($$ == -1) YYERROR; }
	|   '-' expr		    %prec UMINUS
		{ $$ = ExprUnary (compile, ExprUminus, $2); if ($$ == -1) YYERROR; }
	|   '!' expr
		{ $$ = ExprUnary (compile, ExprNot, $2); if ($$ == -1) YYERROR; }
	|   '~' expr
		{ $$ = ExprUnary (compile, ExprLnot, $2); if ($$ == -1) YYERROR; }
	|   expr ',' expr
		{
		    $$ = ExprTree (compile, ExprComma, $1, $3); 
		    if ($$ == -1) YYERROR;
		    CompExpr (compile, $$)->nvalues = 
			CompExpr (compile, $1)->nvalues +
			CompExpr (compile, $3)->nvalues;
		}
	|   '(' expr ')'
		{ $$ = $2; }
	|   NAME '(' oexpr ')'
		{   
		    FormInt	func, nparam, nresult;
		    FormBuiltin	*b;
		    FormFunc	*f;

		    
		    if ($1 == -1)
			YYERROR;
		    if ($3 == -1)
			nparam = 0;
		    else
			nparam = CompExpr (compile, $3)->nvalues;
		    if ((func = FuncLookup (compile, $1)) != -1)
		    {
			f = CompFunc (compile, func);
			if (f->numParam != nparam)
			{
			    yyerror ("%s wants %d given %d",
				     CompString (compile,
						 f->name),
				     f->numParam, nparam);
			    YYERROR;
			}
			$$ = ExprTree (compile, ExprCall,
				       ExprVar (compile, func), $3); 
			if ($$ == -1) 
			    YYERROR;
			nresult = f->numResult;
			if (nresult == -1)
			    nresult = f->numParam;
			CompExpr (compile, $$)->nvalues = nresult;
		    } 
		    else if ((b = (*builtinLookup) (compile, $1)))
		    {
			if (b->numParam != nparam)
			{
			    yyerror ("%s wants %d given %d",
				     b->name, b->numParam, nparam);
			    YYERROR;
			}
			$$ = ExprTree (compile, ExprBuiltin,
				       ExprBuilt (compile, b->id), $3);
			CompExpr (compile, $$)->nvalues = b->numResult;
		    }
		    else
		    {
			yyerror ("undefined \"%s\"",
				 CompString (compile, $1));
			YYERROR;
		    }
		}
	;
oexpr	:   expr
	|
		{ $$ = -1; }
	;
%%

static int
yylex(void)
{
    int	    r;
    char    token[LEX_TOKEN];

    token[0] = '\0';
    r = (*lexp)(token);
    switch (r) {
    case NAME:
    case LABEL:
	yylval = StringAlloc (compile, token);
	break;
    case NUMBER:
	yylval = ConstAlloc (compile, yydval);
	break;
    }
    return r;
}

int
yywrap (void)
{
    return 1;
}

#include    <stdarg.h>

extern int  lexline, lexchar;

static void (*yyvsprint) (char *, char *, va_list args) = vsprint;
static unsigned char *(*lexgetlinep) (int *) = lexgetline;

void
yysprint (char *dst, char *format, ...)
{
    va_list	args;
    va_start (args, format);
    yyvsprint (dst, format, args);
    va_end (args);
}

void
yyerror (char *fmt, ...)
{
    char	*d;
    va_list	args;
    unsigned char   *line;
    int		    len;
    int		    c;
    
    d = compile->error;
    va_start (args, fmt);
    yysprint (d, "Line %d: ", lexline);
    d = d + StrLen (d);
    yyvsprint (d, fmt, args);
    d = d + StrLen (d);
    *d++ = '\n';
    line = (*lexgetlinep) (&len);
    c = lexchar;
    while (len && d < compile->error + BUF_ERROR - 3)
    {
	*d++ = *line++;
	len--;
	c--;
	if (c == 0)
	    *d++ = '@';
    }
    if (c)
	*d++ = '@';
    *d++ = '\0';
    va_end (args);
}

void
yyinit (char *program, int programSize)
{
    yylexinit (program, programSize);
    MemSet (compile, sizeof (Compile), '\0');
#ifdef YYDEBUG
    yydebug = 1;
#endif
}

FormPickle *
FormPack (Compile *c)
{
    FormPickle	*p;
    FormInt	nstorage;
    FormByte	*storage;
    FormNumber	*consts;
    FormVar	*vars;
    FormFunc	*funcs;
    FormInt	*globals;
    FormByte	*code;
    FormByte	*strings;
    int		i;
    
    nstorage = (sizeof (FormPickle) +
		c->nconst * sizeof (FormNumber) +
		c->nglobal * sizeof (FormInt) +
		c->nvar * sizeof (FormVar) +
		c->nfunc * sizeof (FormFunc) +
		c->nstringbuf +
		c->nexe);
    storage = (FormByte *) MemPtrNew (nstorage);
    if (!storage)
	return 0;
    
    p = (void *) storage;
    storage = storage + sizeof (FormPickle);
    
    consts = (void *) storage;
    for (i = 0; i < c->nconst; i++)
	consts[i] = c->consts[i];
    storage = storage + c->nconst * sizeof (FormNumber);
    
    globals = (void *) storage;
    for (i = 0; i < c->nglobal; i++)
	globals[i] = c->globals[i];
    storage = storage + c->nglobal * sizeof (FormInt);
	
    vars = (void *) storage;
    for (i = 0; i < c->nvar; i++)
	vars[i] = c->var[i].var;
    storage = storage + c->nvar * sizeof (FormVar);
    
    funcs = (void *) storage;
    for (i = 0; i < c->nfunc; i++)
	funcs[i] = c->funcs[i];
    storage = storage + c->nfunc * sizeof (FormFunc);
    
    code = storage;
    MemMove (code, c->exebuf, c->nexe);
    storage = storage + c->nexe;
    
    strings = storage;
    MemMove (strings, c->stringbuf, c->nstringbuf);
    storage = storage + c->nstringbuf;
    
    p->name = c->name + (strings - (FormByte *) p);
    p->version = FormulaOpVersion;
    p->numVar = c->nvar;
    p->numFunc = c->nfunc;
    p->numConst = c->nconst;
    p->numGlobal = c->nglobal;
    p->codeSize = c->nexe;
    p->stringSize = c->nstringbuf;
    p->totalSize = nstorage;
    return p;
}

FormPickle *
FormParse (char *program, int programSize)
{
    FormPickle	*p = 0;
    
    compile = (Compile*) MemPtrNew (sizeof (Compile));
    if (!compile)
    {
	(*formError) ("Can't allocate compile space %d", (int) sizeof (Compile));
	return 0;
    }
    yyinit (program, programSize);
    if (!yyparse ())
    {
	FreeTemp (compile);
	p = FormPack (compile);
	FreeHandles (compile);
    }
    else
    {
	FreeHandles (compile);
	(*formError) ("%s", compile->error);
    }
    MemPtrFree (compile);
    return p;
}
