/*
 * $Id: node.c,v 1.9 2000/09/07 07:10:06 keithp Exp $
 *
 * Copyright  2000 Keith Packard
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Keith Packard not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  Keith Packard makes no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * KEITH PACKARD DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL KEITH PACKARD BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

#include    "formula.h"

#if 0
static int  (*printfp)(const char *, ...) = printf;

#define printf (*printfp)
#endif

FormInt
ExprAlloc (Compile  *c, ExprType type)
{
    FormInt i;
    if (c->nexpr == c->maxexpr)
    {
	if (!ReallocTable (c, &c->exprHand, 
			   (c->maxexpr + NUM_EXPR) * sizeof (Expr)))
	{
	    yyerror ("overflow expr table");
	    return -1;
	}
	c->maxexpr += NUM_EXPR;
    }
    i = c->nexpr++;
    c->expr[i].type = type;
    c->expr[i].nvalues = 0;
    return i;
}

FormInt
ExprTree (Compile *c, ExprType type, FormInt left, FormInt right)
{
    FormInt e;

    e = ExprAlloc (c, type);
    if (e == -1)
	return -1;
    CompExpr (c, e)->u.t.left = left;
    CompExpr (c, e)->u.t.right = right;
    return e;
}

FormInt
ExprBinary (Compile *c, ExprType type, FormInt left, FormInt right)
{
    FormInt e;

    e = ExprTree (c, type, left, right);
    if (CompExpr (c, left)->nvalues != 1 ||
	CompExpr (c, right)->nvalues != 1)
    {
	yyerror ("expr has more than one value");
	return -1;
    }
    CompExpr (c, e)->nvalues = 1;
    return e;
}

FormInt
ExprUnary (Compile *c, ExprType type, FormInt down)
{
    FormInt e;

    e = ExprTree (c, type, down, -1);
    if (CompExpr (c, down)->nvalues != 1)
    {
	yyerror ("expr has more than one value");
	return -1;
    }
    CompExpr (c, e)->nvalues = 1;
    return e;
}

FormInt
ExprVar (Compile *c, FormInt name)
{
    FormInt e;

    e = ExprAlloc (c, ExprName);
    if (e == -1)
	return -1;
    CompExpr (c, e)->u.name = name;
    CompExpr (c, e)->nvalues = 1;
    return e;
}

FormInt
ExprVarCheck (Compile *c, FormInt name)
{
    FormInt e;

    e = ExprAlloc (c, ExprCheck);
    if (e == -1)
	return -1;
    CompExpr (c, e)->u.name = name;
    CompExpr (c, e)->nvalues = 1;
    return e;
}

FormInt
ExprBuilt (Compile *c, FormInt id)
{
    FormInt e;

    e = ExprAlloc (c, ExprName);
    if (e == -1)
	return -1;
    CompExpr (c, e)->u.builtin = id;
    CompExpr (c, e)->nvalues = 1;
    return e;
}

FormInt
ExprNumber (Compile *c, FormInt number)
{
    FormInt e;

    e = ExprAlloc (c, ExprConst);
    if (e == -1)
	return -1;
    CompExpr (c, e)->u.number = number;
    CompExpr (c, e)->nvalues = 1;
    return e;
}

FormInt
ConstAlloc (Compile *c, FormNumber n)
{
    FormInt i;

    for (i = 0; i < c->nconst; i++)
	if (*(unsigned long *)&c->consts[i] == *(unsigned long *)&n)
	    return i;
    if (i == c->maxconst)
    {
	if (!ReallocTable (c, &c->constsHand,
			   (c->maxconst + NUM_CONST) * sizeof (FormNumber)))
	{
	    yyerror ("overflow const table");
	    return -1;
	}
	c->maxconst += NUM_CONST;
    }
    c->consts[i] = n;
    c->nconst = i + 1;
    return i;
}

FormInt
NameAlloc (Compile *c, FormInt name, FormInt optional, FormInt label, FormClass class)
{
    FormInt i;
    
    for (i = c->nvar; --i >= 0;)
    {
	if (!c->var[i].scopedOut &&
	    c->var[i].var.name == name &&
	    ((c->var[i].var.class == FormClassGlobal) == 
	     (class == FormClassGlobal)))
	{
	    yyerror ("redefinition of name %s", CompString (c, name));
	    return -2;
	}
    }
    if (c->nvar == c->maxvar)
    {
	if (!ReallocTable (c, &c->varHand,
			   (c->maxvar + NUM_VAR) * sizeof (Name)))
	{
	    yyerror ("overflow var table");
	    return -1;
	}
	c->maxvar += NUM_VAR;
    }
    i = c->nvar++;
    c->var[i].var.name = name;
    c->var[i].var.label = label;
    c->var[i].var.class = class;
    c->var[i].var.optional = optional;
    c->var[i].scopedOut = 0;
    return i;
}

FormInt
NameLookup (Compile *c, FormInt name)
{
    FormInt i;

    for (i = c->nvar; --i >= 0;)
    {
	if (!c->var[i].scopedOut &&
	    c->var[i].var.name == name)
	    return i;
    }
    return -1;
}

#if 0
FormInt
ArrayAlloc (Compile *c, FormInt dim, FormInt array)
{
    FormInt	i;
    FormArray	*a;

    if (c->narray == c->maxarray)
    {
	if (!Realloctable (c, &c->arrayHand,
			   (c->maxarray + NUM_ARRAY) * sizeof (FormArray)))
	{
	    yyerror ("overflow array table");
	    return -1;
	}
	c->maxarray += NUM_ARRAY;
    }
    i = c->narray++;
    a = CompArray(c, i);
    f->dim = dim;
    f->array = array;
    return i;
}
#endif

FormInt
FuncAlloc (Compile *c, FormInt name)
{
    FormFunc	*f;
    FormInt	i;

    for (i = 0; i < c->nfunc; i++)
	if (c->funcs[i].name == name)
	{
	    yyerror ("redefinition of function %s", CompString (c, name));
	    return -1;
	}
    if (c->nfunc == c->maxfunc)
    {
	if (!ReallocTable (c, &c->funcsHand,
			   (c->maxfunc + NUM_FUNC) * sizeof (FormFunc)))
	{
	    yyerror ("overflow func table");
	    return -1;
	}
	c->maxfunc += NUM_FUNC;
    }
    i = c->nfunc++;
    f = CompFunc(c,i);
    f->name = name;
    f->numParam = 0;
    f->numResult = 0;
    f->numLocal = 0;
    return i;
}
    
FormInt
FuncLookup (Compile *c, FormInt name)
{
    FormInt i;

    for (i = 0; i < c->nfunc; i++)
	if (c->funcs[i].name == name)
	    return i;
    return -1;
}

FormInt
GlobalAlloc (Compile *c, FormInt n)
{
    FormInt g;
    
    if (c->nglobal == c->maxglobal)
    {
	if (!ReallocTable (c, &c->globalsHand,
			   (c->maxglobal + NUM_GLOBAL) * sizeof (FormInt)))
	{
	    yyerror ("overflow global table");
	    return -1;
	}
	c->maxglobal += NUM_GLOBAL;
    }
    g = compile->nglobal++;
    *CompGlobal (c, g) = n;
    return g;
}

FormInt
StringAlloc (Compile *c, FormByte *string)
{
    FormInt	i;
    FormInt	len;
    FormInt	size;

    for (i = 0; i < c->nstring; i++)
    {
	if (!StrCompare (string, CompString (c, c->strings[i])))
	    return c->strings[i];
    }
    if (c->nstring == c->maxstring)
    {
	if (!ReallocTable (c, &c->stringsHand,
			   (c->maxstring + NUM_STRING) * sizeof (FormInt)))
	{
	    yyerror ("overflow string table");
	    return -1;
	}
	c->maxstring += NUM_STRING;
    }
    len = (FormInt) StrLen ((char *) string) + 1;
    size = c->nstringbuf + len;
    if (size > c->maxstringbuf)
    {
	if (size < c->maxstringbuf + BUF_STRING)
	    size = c->maxstringbuf + BUF_STRING;
	if (!ReallocTable (c, &c->stringbufHand, size))
	{
	    yyerror ("overflow string buffer");
	    return -1;
	}
	c->maxstringbuf = size;
    }
    i = c->nstring++;
    c->strings[i] = c->nstringbuf;
    StrCopy ((char *) CompString(c,c->strings[i]), (char *) string);
    c->nstringbuf += len;
    return c->strings[i];
}
	     
void
CleanLocals (Compile *c)
{
    FormInt i;

    for (i = 0; i < c->nvar; i++)
    {
	if (!c->var[i].scopedOut &&
	    c->var[i].var.class != FormClassGlobal)
	{
	    c->var[i].scopedOut = 1;
	}
    }
}

/*
 * Save the current location on the label stack
 */

FormInt
PushLabel (Compile *c, LabelType type, FormInt e)
{
    FormInt l;

    l = c->nlabel;
    if (l == c->maxlabel)
    {
	if (!ReallocTable (c, &c->labelsHand,
			   (c->maxlabel + NUM_LABEL) * sizeof (Label)))
	{
	    yyerror ("overflow branch table");
	    return -1;
	}
	c->maxlabel += NUM_LABEL;
    }
    c->nlabel++;
    CompLabel(c,l)->exe = e;
    CompLabel(c,l)->type = type;
    return l;
}

FormInt
NewForwardBranch (Compile *c, LabelType type)
{
    FormInt l;
    
    l = PushLabel (c, type, -1);
    if (l >= 0)
    {
	CompLabel (c, l)->reverse = 0;
    }
    return l;
}

FormInt
UseBranch (Compile *c, LabelType type, FormOpCode op)
{
    FormInt l;
    FormInt e;

    for (l = c->nlabel; l--;)
	if (CompLabel(c,l)->type == type)
	    break;
    if (l >= 0)
    {
	if (!CompileOp (c, op))
	    return -1;
	e = c->nexe;
        if (!CompileInt (c, CompLabel (c, l)->exe))
	    return -1;
	if (!CompLabel(c,l)->reverse)
	    CompLabel (c, l)->exe = e;
    }
    return l;
}

FormInt
ForwardBranchDest (Compile *c, LabelType type)
{
    FormInt l;
    FormInt e, next;

    for (l = c->nlabel; l--;)
	if (CompLabel(c,l)->type == type)
	    break;
    if (l >= 0)
    {
	e = CompLabel(c,l)->exe;
	while (e != -1)
	{
	    next = CompileFetchInt (c, e);
	    CompilePatchInt (c, e, c->nexe);
	    e = next;
	}
    }
    return l;
}

FormInt
ReverseBranchDest (Compile *c, LabelType type)
{
    FormInt l;

    l = PushLabel (c, type, c->nexe);
    if (l >= 0)
	CompLabel (c, l)->reverse = 1;
    return l;
}

void
CleanLabel (Compile *c, FormInt n)
{
    c->nlabel -= n;
    if (c->nlabel < 0)
	c->nlabel = 0;
}

void
LockHandles (Compile *c)
{
    if (c->varHand)
	c->var = MemHandleLock (c->varHand);
#if 0
    if (c->arrayHand)
	c->array = MemHandleLock (c->arrayHand);
#endif
    if (c->globalsHand)
	c->globals = MemHandleLock (c->globalsHand);
    if (c->constsHand)
	c->consts = MemHandleLock (c->constsHand);
    if (c->exprHand)
	c->expr = MemHandleLock (c->exprHand);
    if (c->funcsHand)
	c->funcs = MemHandleLock (c->funcsHand);
    if (c->stringsHand)
	c->strings = MemHandleLock (c->stringsHand);
    if (c->stringbufHand)
	c->stringbuf = MemHandleLock (c->stringbufHand);
    if (c->labelsHand)
	c->labels = MemHandleLock (c->labelsHand);
    if (c->exebufHand)
	c->exebuf = MemHandleLock (c->exebufHand);
}

void
UnlockHandles (Compile *c)
{
    if (c->varHand)
	MemHandleUnlock (c->varHand);
#if 0
    if (c->arrayHand)
	MemHandleUnlock (c->arrayHand);
#endif
    if (c->globalsHand)
	MemHandleUnlock (c->globalsHand);
    if (c->constsHand)
	MemHandleUnlock (c->constsHand);
    if (c->exprHand)
	MemHandleUnlock (c->exprHand);
    if (c->funcsHand)
	MemHandleUnlock (c->funcsHand);
    if (c->stringsHand)
	MemHandleUnlock (c->stringsHand);
    if (c->stringbufHand)
	MemHandleUnlock (c->stringbufHand);
    if (c->labelsHand)
	MemHandleUnlock (c->labelsHand);
    if (c->exebufHand)
	MemHandleUnlock (c->exebufHand);
}

void
FreeTemp (Compile *c)
{
    if (c->exprHand)
    {
	MemHandleUnlock (c->exprHand);
	MemHandleFree (c->exprHand);
	c->exprHand = 0;
    }
    if (c->labelsHand)
    {
	MemHandleUnlock (c->labelsHand);
	MemHandleFree (c->labelsHand);
	c->labelsHand = 0;
    }
}
void
FreeHandles (Compile *c)
{
    UnlockHandles (c);
    if (c->varHand)
	MemHandleFree (c->varHand);
#if 0
    if (c->arrayHand)
	MemHandleFree (c->arrayHand);
#endif
    if (c->globalsHand)
	MemHandleFree (c->globalsHand);
    if (c->constsHand)
	MemHandleFree (c->constsHand);
    if (c->exprHand)
	MemHandleFree (c->exprHand);
    if (c->funcsHand)
	MemHandleFree (c->funcsHand);
    if (c->stringsHand)
	MemHandleFree (c->stringsHand);
    if (c->stringbufHand)
	MemHandleFree (c->stringbufHand);
    if (c->labelsHand)
	MemHandleFree (c->labelsHand);
    if (c->exebufHand)
	MemHandleFree (c->exebufHand);
}

FormInt
ReallocTable (Compile	*c,
	      VoidHand	*h,
	      FormInt	size)
{
    FormInt r = 0;
    
    UnlockHandles (c);
    if (*h)
    {
	if (MemHandleResize (*h, size) == 0)
	    r = 1;
    }
    else
    {
	*h = MemHandleNew (size);
	if (*h)
	    r = 1;
    }
    LockHandles (c);
    return r;
}
