#include    <stdio.h>
#include    "formula.h"

FormNumber
MachFetch (FormMachine	*m,
	   FormInt	sym)
{
    FormVar	*v = &m->exe.vars[sym];
    FormNumber	a;

    switch (v->class) {
    case FormClassParam:
	a = m->value[m->fp + v->location];
	break;
    case FormClassLocal:
    case FormClassResult:
	a = m->value[m->fp - (v->location + 1)];
	break;
    case FormClassStatic:
	a = 0;
    }
    printf ("Fetch %s %g\n", m->exe.strings + v->name, a);
    return a;
}

void
PrintInst (FormMachine	*m)
{
    char	*s;
    FormByte	*c;
    
    c = GetCode (&m->exe, m->pc);
    switch (*c) {
    case OpNoop: s = "Noop"; break;
    case OpPush: s = "Push"; break;
    case OpPop: s = "Pop"; break;
    case OpJmp: s = "Jmp"; break;
    case OpJne: s = "Jne"; break;
    case OpJeq: s = "Jeq"; break;
    case OpReturn: s = "Return"; break;
    case OpCall: s = "Call"; break;
    case OpNameGet: s = "NameGet"; break;
    case OpNamePut: s = "NamePut"; break;
    case OpConst: s = "Const"; break;
    case OpUminus: s = "Uminus"; break;
    case OpLnot: s = "Lnot"; break;
    case OpBang: s = "Bang"; break;
    case OpFact: s = "Fact"; break;
    case OpPlus: s = "Plus"; break;
    case OpMinus: s = "Minus"; break;
    case OpTimes: s = "Times"; break;
    case OpDivide: s = "Divide"; break;
    case OpDiv: s = "Div"; break;
    case OpMod: s = "Mod"; break;
    case OpPow: s = "Pow"; break;
    case OpLxor: s = "Lxor"; break;
    case OpLand: s = "Land"; break; 
    case OpLor: s = "Lor"; break;
    case OpAnd: s = "And"; break; 
    case OpOr: s = "Or"; break;
    case OpEq: s = "Eq"; break;
    case OpNe: s = "Ne"; break;
    case OpLt: s = "Lt"; break;
    case OpGt: s = "Gt"; break;
    case OpLe: s = "Le"; break;
    case OpGe: s = "Ge"; break;
    case OpEnd: s = "End"; break;
    }
    printf ("%s\n", s);
}

void
MachStore (FormMachine	*m,
	   FormInt	sym,
	   FormNumber	a)
{
    FormVar	*v = &m->exe.vars[sym];

    printf ("Store %s %g\n", m->exe.strings + v->name, a);
    switch (v->class) {
    case FormClassParam:
	m->value[m->fp + v->location] = a;
	break;
    case FormClassLocal:
    case FormClassResult:
	m->value[m->fp - (v->location + 1)] = a;
	break;
    case FormClassStatic:
	break;
    }
}

FormInt
MachInline (FormMachine *m)
{
    FormByte	*c;
    FormInt	i;

    c = GetCode (&m->exe, m->pc);
    if (m->exe.model == FormModelSmall)
    {
	i = c[0] | (c[1] << 8);
	m->pc += 2;
    }
    else
    {
	i = c[0] | (c[1] << 8) | (c[2] << 16) | (c[3] << 24);
	m->pc += 4;
    }
    return i;
}

FormInt
MachRun (FormMachine	*m)
{
    FormByte	*c;
    FormInt	i;
    FormInt	running = 1;
    
    while (running)
    {
	PrintInst (m);
	c = GetCode (&m->exe, m->pc);
	m->pc++;
	switch (*c) {
	case OpNoop:
	    break;
	case OpPush:
	    ValuePush (m, m->a); break;
	case OpPop:
	    m->a = ValuePop (m); break;
	case OpJmp:
	    m->pc = MachInline (m);
	    break;
	case OpJne:
	    i = MachInline (m);
	    if (m->a != 0)
		m->pc = i;
	    break;
	case OpJeq:
	    i = MachInline (m);
	    if (m->a == 0)
		m->pc = i;
	    break;
	case OpReturn:
	    break;
	case OpCall:
    /*	MachCall (m, MachInline (m)); */
	    break;
	case OpNameGet:
	    m->a = MachFetch (m, MachInline (m));
	    break;
	case OpNamePut:
	    MachStore (m, MachInline (m), m->a);
	    break;
	case OpConst:
	    m->a = m->exe.consts[MachInline (m)];
	    break;
	case OpUminus:
	    m->a = -m->a; break;
	case OpLnot:
	    m->a = ~(int) m->a; break;
	case OpBang:
	    m->a = !m->a; break;
	case OpFact:
    /*	m->a = MachFactorial (m->a); */ break; 
	case OpPlus:
	    m->a = ValuePop (m) + m->a; break;
	case OpMinus:
	    m->a = ValuePop (m) - m->a; break;
	case OpTimes:
	    m->a = ValuePop (m) * m->a; break;
	case OpDivide:
	    m->a = ValuePop (m) / m->a; break;
	case OpDiv:
	    m->a = (int) ValuePop (m) / (int) m->a; break;
	case OpMod:
	    m->a = (int) ValuePop (m) % (int) m->a; break;
	case OpPow:
    /*	m->a = MachPow (ValuePop (m), m->a); */ break;
	case OpLxor:
	    m->a = (int) ValuePop (m) ^ (int) m->a; break;
	case OpLand:
	    m->a = (int) ValuePop (m) & (int) m->a; break;
	case OpLor:
	    m->a = (int) ValuePop (m) | (int)  m->a; break;
	case OpEq:
	    m->a = ValuePop (m) == m->a; break;
	case OpNe:
	    m->a = ValuePop (m) != m->a; break;
	case OpLt:
	    m->a = ValuePop (m) < m->a; break;
	case OpGt:
	    m->a = ValuePop (m) > m->a; break;
	case OpLe:
	    m->a = ValuePop (m) <= m->a; break;
	case OpGe:
	    m->a = ValuePop (m) >= m->a; break;
	case OpEnd:
	    running = 0; break;
	}
    }
    return running;
}
