typedef double		FormNumber;
typedef unsigned char	FormByte;
typedef short		FormInt;
typedef short		FormString;

typedef enum _OpCode {
    OpNoop,
    OpPush,
    OpPop,
    /*
     * control flow op codes.  2-byte argument: signed offset in bytes
     */
    OpJmp,
    OpJne,
    OpJeq,
    /*
     * function op codes
     */
    OpReturn,	/* no argument */
    OpCall,	/* 1-byte argument -- index into global table */
    /*
     * terminal op codes.  1-byte argument: index into global table
     */
    OpNameGet,
    OpNamePut,
    OpConst,
    /*
     * operator op codes.  No argument
     */
    OpUminus,
    OpLnot,
    OpBang,
    OpFact,
    OpPlus,
    OpMinus,
    OpTimes,
    OpDivide,
    OpDiv,
    OpMod,
    OpPow,
    OpLxor,
    OpLand, 
    OpLor,
    OpAnd, 
    OpOr,
    OpEq,
    OpNe,
    OpLt,
    OpGt,
    OpLe,
    OpGe,
    /*
     * last op code
     */
    OpEnd,
} FormOpCode;     

/*
 * Format of executable:
 *
 * Header
 *  name
 *  model   (small, medium, large)
 *  number of variables
 *  number of functions
 *  number of constants
 *  size of code block
 *  size of string table
 * Variables
 *  name
 *  label
 *  class
 *  location
 * Functions
 *  name
 *  label
 *  number parameters
 *  number results
 *  number locals
 *  code size
 *  offset of code block
 * Constants
 * Code Blocks
 *  Parameters
 *   list of variables
 *  Results
 *   list of variables
 *  Locals
 *   list of variables
 *  code
 *   list of bytes
 * Strings
 *   list of bytes; each string is null-terminated
 */

typedef enum _FormModel {
    FormModelSmall, FormModelLarge
} FormModel;

typedef enum _FormClass {
    FormClassParam,
    FormClassResult,
    FormClassLocal,
    FormClassStatic
} FormClass;

typedef struct _FormVar {
    FormString	name;
    FormString	label;
    FormClass	class;
    FormInt	location;
} FormVar;

typedef struct _FormFunc {
    FormString	name;
    FormString	label;
    FormInt	numParam;
    FormInt	numResult;
    FormInt	numLocal;
    FormInt	codeSize;
    FormInt	codeBlock;
} FormFunc;

typedef struct _FormExe {
    FormString	name;
    FormModel	model;
    FormInt	numVar;
    FormVar	*vars;
    FormInt	numFunc;
    FormFunc	*funcs;
    FormInt	numConst;
    FormNumber	*consts;
    FormByte	*code;
    FormByte	*strings;
    FormByte	*storage;
} FormExe;

#define GetString(e,n)	((e)->strings + (n))
#define GetConst(e,n)	((e)->consts[n])
#define GetCode(e,n)	((e)->code + (n))

#define VALUE_STACK 256
#define RET_STACK   256

typedef struct _FormMachine {
    FormExe	exe;
    FormInt	pc;
    FormInt	sp;
    FormInt	fp;
    FormInt	rp;
    FormNumber	a;
    FormNumber	value[VALUE_STACK];
    FormInt	ret[RET_STACK];
} FormMachine;

typedef FormByte (*FormReader)(FormInt);

#define ValuePush(m,v)	((m)->value[--(m)->sp] = (v))
#define ValuePop(m)	((m)->value[(m)->sp++])
#define ValueZero(v)	((v) == 0)

FormInt	MachRun (FormMachine *m);

    
